/*
 * Copyright 2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.plugins.relaxNG.compact;

import org.intellij.plugins.relaxNG.compact.folding.RncFoldingBuilder;
import org.intellij.plugins.relaxNG.compact.formatting.RncFormattingModelBuilder;
import org.intellij.plugins.relaxNG.compact.psi.RncDefine;
import org.intellij.plugins.relaxNG.compact.psi.RncElement;
import org.intellij.plugins.relaxNG.compact.psi.util.EscapeUtil;
import org.intellij.plugins.relaxNG.compact.psi.util.RenameUtil;

import com.intellij.formatting.FormattingModelBuilder;
import com.intellij.lang.*;
import com.intellij.lang.annotation.Annotator;
import com.intellij.lang.documentation.DocumentationProvider;
import com.intellij.lang.folding.FoldingBuilder;
import com.intellij.lang.refactoring.NamesValidator;
import com.intellij.openapi.fileTypes.SyntaxHighlighter;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.psi.PsiComment;
import com.intellij.psi.PsiElement;
import com.intellij.psi.PsiManager;
import com.intellij.psi.PsiWhiteSpace;
import com.intellij.psi.tree.IElementType;
import com.intellij.psi.tree.TokenSet;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

/*
* Created by IntelliJ IDEA.
* User: sweinreuter
* Date: 01.08.2007
*/
public class RngCompactLanguage extends Language {
  private static final String ID = "RELAX-NG";

  public static final RngCompactLanguage INSTANCE = new RngCompactLanguage();

  private final RncFoldingBuilder myFoldingBuilder = new RncFoldingBuilder();
  private final RncParserDefinition myParserDefinition = new RncParserDefinition();
  private final MyCommenter myCommenter = new MyCommenter();
  private final MyPairedBraceMatcher myPairedBraceMatcher = new MyPairedBraceMatcher();
  private final ReferenceAnnotator myAnnotator = new ReferenceAnnotator();
  private final RncFormattingModelBuilder myFormattingModelBuilder = new RncFormattingModelBuilder();

  private RngCompactLanguage() {
    super(ID);
  }

  @Override
  @NotNull
  public SyntaxHighlighter getSyntaxHighlighter(Project project, VirtualFile virtualFile) {
    return new RncHighlighter();
  }

  @Nullable
  public Annotator getAnnotator() {
    return myAnnotator;
  }

  @Nullable
  public ParserDefinition getParserDefinition() {
    return myParserDefinition;
  }

  @Override
  @Nullable
  public PairedBraceMatcher getPairedBraceMatcher() {
    return myPairedBraceMatcher;
  }

  @Override
  @Nullable
  public Commenter getCommenter() {
    return myCommenter;
  }

  @Override
  @Nullable
  public FoldingBuilder getFoldingBuilder() {
    return myFoldingBuilder;
  }

  @Nullable
  public FormattingModelBuilder getFormattingModelBuilder() {
    return myFormattingModelBuilder;
  }

  @NotNull
  public NamesValidator getNamesValidator() {
    return new MyNamesValidator();
  }

  @Override
  @NotNull
  public TokenSet getReadableTextContainerElements() {
    return RncTokenTypes.READABLE_TEXT;
  }

  @Override
  @Nullable
  protected DocumentationProvider createDocumentationProvider() {
    return new MyDocumentationProvider();
  }

  private static class MyCommenter implements Commenter {
    @Nullable
    public String getLineCommentPrefix() {
      return "#";
    }

    @Nullable
    public String getBlockCommentPrefix() {
      return null;
    }

    @Nullable
    public String getBlockCommentSuffix() {
      return null;
    }
  }

  private static class MyPairedBraceMatcher implements PairedBraceMatcher {
    private BracePair[] myBracePairs;

    public BracePair[] getPairs() {
      if (myBracePairs == null) {
        myBracePairs = new BracePair[]{
                new BracePair('{', RncTokenTypes.LBRACE, '}', RncTokenTypes.RBRACE, true),
                new BracePair('(', RncTokenTypes.LPAREN, ')', RncTokenTypes.RPAREN, false),
                new BracePair('[', RncTokenTypes.LBRACKET, ']', RncTokenTypes.RBRACKET, false),
        };
      }
      return myBracePairs;
    }

    public boolean isPairedBracesAllowedBeforeType(@NotNull IElementType lbraceType, @Nullable IElementType contextType) {
      return false;
    }
  }

  private static class MyNamesValidator implements NamesValidator {
    public boolean isKeyword(String name, Project project) {
      return RenameUtil.isKeyword(PsiManager.getInstance(project), name);
    }

    public boolean isIdentifier(String name, Project project) {
      return RenameUtil.isIdentifier(PsiManager.getInstance(project), name);
    }
  }

  private static class MyDocumentationProvider implements DocumentationProvider {
    @Nullable
    public String getQuickNavigateInfo(PsiElement element) {
      return null;  //To change body of implemented methods use File | Settings | File Templates.
    }

    @Nullable
    public String getUrlFor(PsiElement element, PsiElement originalElement) {
      return null;
    }

    @Nullable
    public String generateDoc(PsiElement element, PsiElement originalElement) {
      if (element instanceof RncElement) {
        PsiElement comment = element.getPrevSibling();
        while (comment instanceof PsiWhiteSpace) {
          comment = comment.getPrevSibling();
        }
        if (comment instanceof PsiComment) {
          final StringBuilder sb = new StringBuilder();
          do {
            sb.insert(0, EscapeUtil.unescapeText(comment).replaceAll("\n?##?", "") + "<br>");
            comment = comment.getPrevSibling();
          } while (comment instanceof PsiComment);

          if (element instanceof RncDefine) {
            sb.insert(0, "Define: <b>" + ((RncDefine)element).getName() + "</b><br>");
          }
          
          return sb.toString();
        }
      }
      return null;
    }

    @Nullable
    public PsiElement getDocumentationElementForLookupItem(PsiManager psiManager, Object object, PsiElement element) {
      return null;
    }

    @Nullable
    public PsiElement getDocumentationElementForLink(PsiManager psiManager, String link, PsiElement context) {
      return null;
    }
  }
}
