/*
 * Copyright 2007 Sascha Weinreuter
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.intellij.plugins.relaxNG.model.resolve;

import org.intellij.plugins.relaxNG.compact.RncFileType;
import org.intellij.plugins.relaxNG.model.*;
import org.intellij.plugins.relaxNG.xml.dom.RngGrammar;

import com.intellij.ProjectTopics;
import com.intellij.navigation.ItemPresentation;
import com.intellij.navigation.NavigationItem;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.components.ProjectComponent;
import com.intellij.openapi.editor.colors.TextAttributesKey;
import com.intellij.openapi.fileTypes.FileType;
import com.intellij.openapi.fileTypes.StdFileTypes;
import com.intellij.openapi.progress.ProgressIndicator;
import com.intellij.openapi.progress.ProgressManager;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.roots.*;
import com.intellij.openapi.startup.StartupManager;
import com.intellij.openapi.vcs.FileStatus;
import com.intellij.openapi.vfs.*;
import com.intellij.psi.*;
import com.intellij.psi.meta.PsiMetaData;
import com.intellij.psi.meta.PsiMetaOwner;
import com.intellij.psi.meta.PsiPresentableMetaData;
import com.intellij.psi.search.PsiElementProcessor;
import com.intellij.psi.util.CachedValue;
import com.intellij.psi.util.CachedValueProvider;
import com.intellij.psi.xml.XmlFile;
import com.intellij.util.messages.MessageBus;
import com.intellij.util.xml.DomManager;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import gnu.trove.THashSet;
import gnu.trove.TObjectHashingStrategy;
import gnu.trove.TObjectProcedure;

import javax.swing.*;
import java.util.*;

/**
 * Created by IntelliJ IDEA.
 * User: sweinreuter
 * Date: 23.10.2007
 */

/**
 * More or less an adapted copy of an early version of com.intellij.lang.javascript.index.JavaScriptIndex
 */
public final class RngIndexImpl extends RngIndex implements ProjectComponent {
  private static final String BUILDING_RNG_INDEX = "Building Relax-NG indices";

  private final Object myDisposeLock = new Object();
  private final Project myProject;
  private final PsiManager myPsiManager;
  private final MessageBus myMessageBus;
  private final DomManager myDomManager;

  private Map<XmlFile, RngIndexEntry> myOldRngFiles;
  private Map<XmlFile, RngIndexEntry> myRngFiles = new LinkedHashMap<XmlFile, RngIndexEntry>();

  private MyTreeChangeListener myTreeChangeListener;
  private VirtualFileListener myFileListener;
  private Runnable myUpdateRunnable;

  RngIndexImpl(Project project, PsiManager psiManager, DomManager domManager, MessageBus messageBus) {
    myProject = project;
    myPsiManager = psiManager;
    myMessageBus = messageBus;
    myDomManager = domManager;
  }

  private boolean acceptFile(VirtualFile fileOrDir) {
    final FileType type = fileOrDir.getFileType();
    if (type != StdFileTypes.XML) {
      return type == RncFileType.INSTANCE;
    }

    final PsiFile psiFile = myPsiManager.findFile(fileOrDir);
    return psiFile != null && isRngFile(psiFile);
  }

  private boolean isRngOrRncFile(@NotNull PsiFile psiFile) {
    final FileType type = psiFile.getFileType();
    if (type != StdFileTypes.XML) {
      return type == RncFileType.INSTANCE;
    }
    return isRngFile(psiFile);
  }

  private boolean isRngFile(PsiFile psiFile) {
    try {
      return myDomManager.getFileElement((XmlFile)psiFile, RngGrammar.class) != null;
    } catch (ClassCastException e) {
      return false; // fileType == XML && !instanceof XmlFile
    }
  }

  public void projectOpened() {
    myFileListener = new VirtualFileAdapter() {
      @Override
      public void fileCreated(VirtualFileEvent event) {
        fileAdded(event.getFile());
      }

      @Override
      public void beforeFileDeletion(VirtualFileEvent event) {
        final VirtualFile fileOrDir = event.getFile();
        if (acceptFile(fileOrDir)) {
          processFileRemoved((XmlFile)myPsiManager.findFile(fileOrDir));
        }
      }
    };
    final ModuleRootListener rootListener = new ModuleRootListener() {
      public void beforeRootsChange(ModuleRootEvent event) {
      }

      public void rootsChanged(ModuleRootEvent event) {
        Runnable runnable = new Runnable() {
          public void run() {
            myOldRngFiles = myRngFiles;
            myRngFiles = new LinkedHashMap<XmlFile, RngIndexEntry>(myOldRngFiles.size());

            if (ApplicationManager.getApplication().isUnitTestMode()) {
              myUpdateRunnable.run();
            } else {
              ProgressManager.getInstance().runProcessWithProgressSynchronously(
                      myUpdateRunnable,
                      BUILDING_RNG_INDEX,
                      false,
                      myProject
              );
            }

            myOldRngFiles = null;
          }
        };
        ApplicationManager.getApplication().invokeLater(runnable);
      }
    };
    myUpdateRunnable = new Runnable() {
      public void run() {
        final ProgressIndicator progress = ProgressManager.getInstance().getProgressIndicator();
        if (progress != null) {
          progress.setIndeterminate(true);
          progress.setText(BUILDING_RNG_INDEX);
          progress.setText2("");
        }
        final ProjectFileIndex fileIndex = ProjectRootManager.getInstance(myProject).getFileIndex();
        final List<VirtualFile> filesToProcess = new ArrayList<VirtualFile>(5);

        fileIndex.iterateContent(new ContentIterator() {
          public boolean processFile(VirtualFile fileOrDir) {
            if (acceptFile(fileOrDir)) {
              filesToProcess.add(fileOrDir);
            }
            return true;
          }
        });

        if (progress != null) {
          progress.setIndeterminate(false);
        }

        int processed = 0;
        for (VirtualFile f : filesToProcess) {
          fileAdded(f);
          ++processed;
          if (progress != null) {
            progress.setFraction((double)processed / filesToProcess.size());
          }
        }
      }
    };
    myTreeChangeListener = new MyTreeChangeListener();

    if (ApplicationManager.getApplication().isUnitTestMode()) {
      myUpdateRunnable.run();
    } else {
      StartupManager.getInstance(myProject).registerPostStartupActivity(new Runnable() {
        // It seems to be pretty pointless to have all those listeners added before project startup is complete
        // (building indices etc.). Especially the ModuleRootListener seems to be invoked way too often during
        // startup.
        public void run() {
          Runnable runnable = new Runnable() {
            public void run() {
              synchronized (myDisposeLock) {
                if (myFileListener != null) {
                  VirtualFileManager.getInstance().addVirtualFileListener(myFileListener);
                  myMessageBus.connect().subscribe(ProjectTopics.PROJECT_ROOTS, rootListener);
                  myPsiManager.addPsiTreeChangeListener(myTreeChangeListener);
                }
              }
            }
          };
          ApplicationManager.getApplication().invokeLater(runnable);

          myUpdateRunnable.run();
        }
      });
    }
  }

  public void projectClosed() {
    VirtualFileManager.getInstance().removeVirtualFileListener(myFileListener);
    myPsiManager.removePsiTreeChangeListener(myTreeChangeListener);
  }

  @NotNull
  @NonNls
  public String getComponentName() {
    return "Relax-NG-Support.Index";
  }

  public void initComponent() {
  }

  public void disposeComponent() {
    synchronized (myDisposeLock) {
      clear();

      myFileListener = null;
      myUpdateRunnable = null;
      myTreeChangeListener = null;
    }
  }

  private void fileAdded(final VirtualFile fileOrDir) {
    if (!fileOrDir.isDirectory() && acceptFile(fileOrDir)) {
      processFileAdded((XmlFile)myPsiManager.findFile(fileOrDir));
    }
  }

  private void processFileAdded(XmlFile psiFile) {
    if (myOldRngFiles != null) {
      final RngIndexEntry indexEntry = myOldRngFiles.get(psiFile);
      if (indexEntry != null) {
        myRngFiles.put(psiFile, indexEntry);
        return;
      }
    }

    final ProgressIndicator progress = ProgressManager.getInstance().getProgressIndicator();
    if (progress != null) {
      final VirtualFile virtualFile = psiFile.getVirtualFile();
      progress.setText2(virtualFile != null ? virtualFile.getPresentableUrl() : psiFile.getName());
    }
    myRngFiles.put(psiFile, new RngIndexEntry(psiFile));
  }

  private void processFileChanged(XmlFile xmlFile) {
    if (!xmlFile.isPhysical()) return;
    final RngIndexEntry indexEntry = myRngFiles.get(xmlFile);
    if (indexEntry == null) {
      processFileAdded(xmlFile);
    }
  }

  private void processFileRemoved(XmlFile xmlFile) {
    myRngFiles.remove(xmlFile);
  }

  public String[] getSymbolNames() {
    final Set<String> symbolNames = new com.intellij.util.containers.HashSet<String>();
    for (RngIndexEntry entry : myRngFiles.values()) {
      entry.fillSymbolNames(symbolNames);
    }
    return symbolNames.toArray(new String[symbolNames.size()]);
  }

  public NavigationItem[] getSymbolsByName(String name) {
    final Set<NavigationItem> symbolNavItems = new com.intellij.util.containers.HashSet<NavigationItem>();
    for (RngIndexEntry entry : myRngFiles.values()) {
      entry.fillSymbolsByName(name, symbolNavItems);
    }
    return symbolNavItems.toArray(new NavigationItem[symbolNavItems.size()]);
  }

  public boolean processForwardDependencies(final XmlFile file, final PsiElementProcessor<XmlFile> processor) {
    final RngIndexEntry entry = myRngFiles.get(file);
    if (entry == null) return false;

    final RngIndexEntry.IndexEntryContent value = entry.myIndexValue.getValue();
    if (value == null) {
      return false;
    }

    return value.myDependencies.forEach(new TObjectProcedure<XmlFile>() {
      public boolean execute(XmlFile xmlFile) {
        if (!processor.execute(xmlFile)) {
          return false;
        }
        return processForwardDependencies(xmlFile, processor);
      }
    });
  }

  public boolean processBackwardDependencies(final XmlFile file, final PsiElementProcessor<XmlFile> processor) {
    for (Map.Entry<XmlFile, RngIndexEntry> entry : myRngFiles.entrySet()) {
      final XmlFile xmlFile = entry.getKey();
      if (isReachableFrom(file, xmlFile)) {
        if (!processor.execute(xmlFile)) {
          return false;
        }
      }
    }
    return true;
  }

  public boolean isReachableFrom(XmlFile which, XmlFile from) {
    final RngIndexEntry target = myRngFiles.get(which);
    return target != null && _isReachableFrom(which, from);
  }

  private boolean _isReachableFrom(XmlFile which, XmlFile from) {
    if (which == from) return false;

    final RngIndexEntry source = myRngFiles.get(from);
    if (source == null) return false;

    final RngIndexEntry.IndexEntryContent content = source.myIndexValue.getValue();
    if (content != null) {
      final Set<XmlFile> dependencies = content.myDependencies;
      if (dependencies.contains(which)) {
        return true;
      }
      for (XmlFile dependency : dependencies) {
        if (_isReachableFrom(dependency, which)) {
          return true;
        }
      }
    }
    return false;
  }

  public void clear() {
    myRngFiles.clear();
  }

  private class MyTreeChangeListener extends PsiTreeChangeAdapter {
    public void childAdded(PsiTreeChangeEvent event) {
      final PsiElement child = event.getChild();
      if (child instanceof PsiFile && child.isPhysical() && isRngOrRncFile((PsiFile)child)) {
        processFileAdded((XmlFile)child);
      } else {
        process(event);
      }
    }

    public void childrenChanged(PsiTreeChangeEvent event) {
      process(event);
    }

    public void childRemoved(PsiTreeChangeEvent event) {
      if (event.getChild() instanceof XmlFile) {
        processFileRemoved((XmlFile)event.getChild());
      } else {
        process(event);
      }
    }

    public void childReplaced(PsiTreeChangeEvent event) {
      process(event);
    }

    private void process(PsiTreeChangeEvent event) {
      final PsiElement element = event.getParent();
      if (element != null) {
        final PsiFile psiFile = element.getContainingFile();
        if (psiFile != null && isRngOrRncFile(psiFile)) {
          processFileChanged((XmlFile)psiFile);
        }
      }
    }
  }

  static class RngIndexEntry {

    private final XmlFile myPsiFile;

    static class IndexEntryContent {
      private final Map<String, Object> mySymbols = new HashMap<String, Object>();
      @SuppressWarnings({ "unchecked" })
      private final THashSet<XmlFile> myDependencies = new THashSet<XmlFile>(TObjectHashingStrategy.IDENTITY);

      void addSymbol(String nameComponent, Define element) {
        final Object o = mySymbols.get(nameComponent);
        if (o == null) {
          mySymbols.put(nameComponent, element);
        } else if (o instanceof Define) {
          mySymbols.put(nameComponent, new Object[]{ o, element });
        } else {
          Object[] oArray = (Object[])o;
          Object[] newArray = new Object[oArray.length + 1];
          System.arraycopy(oArray, 0, newArray, 0, oArray.length);
          newArray[oArray.length] = element;
          mySymbols.put(nameComponent, newArray);
        }
      }

      public void addReferencedFile(XmlFile ref) {
        myDependencies.add(ref);
      }
    }

    private final CachedValue<IndexEntryContent> myIndexValue;

    public RngIndexEntry(XmlFile psiFile) {
      myPsiFile = psiFile;
      myIndexValue = psiFile.getManager().getCachedValuesManager().createCachedValue(new CachedValueProvider<IndexEntryContent>() {
        public Result<IndexEntryContent> compute() {
          return new Result<IndexEntryContent>(collectSymbols(), myPsiFile);
        }

        private IndexEntryContent collectSymbols() {
          final IndexEntryContent content = new IndexEntryContent();

          final Grammar grammar = GrammarFactory.getGrammar(myPsiFile);
          if (grammar != null) {
            grammar.acceptChildren(new CommonElement.Visitor() {
              public void visitGrammar(Grammar pattern) {
                pattern.acceptChildren(this);
              }

              public void visitDiv(Div div) {
                div.acceptChildren(this);
              }

              @Override
              public void visitDefine(Define define) {
                content.addSymbol(define.getName(), define);
                define.acceptChildren(this);
              }

              @Override
              public void visitInclude(Include inc) {
                final PsiFile include = inc.getInclude();
                if (include instanceof XmlFile) {
                  content.addReferencedFile((XmlFile)include);
                }
                inc.acceptChildren(this);
              }
            });
          }
          return content;
        }
      }, false);

      myIndexValue.getValue();
    }

    public void fillSymbolNames(final Set<String> symbolNames) {
      synchronized (this) {
        final IndexEntryContent content = myIndexValue.getValue();
        if (content != null) symbolNames.addAll(content.mySymbols.keySet());
      }
    }

    public void fillSymbolsByName(final String name, final Set<NavigationItem> symbolNavItems) {
      synchronized (this) {
        if (!myPsiFile.isPhysical() && !ApplicationManager.getApplication().isUnitTestMode()) {
          return; // predefined
        }

        final IndexEntryContent index = myIndexValue.getValue();
        if (index != null) {
          final Object value = index.mySymbols.get(name);
          if (value instanceof Object[]) {
            for (Object o : (Object[])value) {
              if (o instanceof NavigationItem) {
                MyNavigationItem.add((NavigationItem)o, symbolNavItems);
              } else if (o instanceof Define) {
                final PsiElement psi = ((Define)o).getPsiElement();
                if (psi != null) {
                  MyNavigationItem.add((NavigationItem)psi, symbolNavItems);
                }
              }
            }
          } else if (value instanceof Define) {
            final Define define = (Define)value;
            final PsiElement psi = define.getPsiElement();
            if (psi != null) {
              MyNavigationItem.add((NavigationItem)define.getPsiElement(), symbolNavItems);
            }
          } else if (value instanceof NavigationItem) {
            MyNavigationItem.add((NavigationItem)value, symbolNavItems);
          }
        }
      }
    }

    public boolean equals(Object o) {
      if (this == o) return true;
      if (o == null || getClass() != o.getClass()) return false;

      final RngIndexEntry that = (RngIndexEntry)o;

      return myPsiFile == that.myPsiFile;
    }

    public int hashCode() {
      return myPsiFile.hashCode();
    }

    private static class MyNavigationItem implements NavigationItem, ItemPresentation {
      private final NavigationItem myItem;
      private final ItemPresentation myPresentation;

      private MyNavigationItem(NavigationItem item, @NotNull final ItemPresentation presentation) {
        myItem = item;
        myPresentation = presentation;
      }

      public String getPresentableText() {
        return myPresentation.getPresentableText();
      }

      @Nullable
      public String getLocationString() {
        return getLocationString((PsiElement)myItem);
      }

      private static String getLocationString(PsiElement element) {
        return "(in " + element.getContainingFile().getName() + ")";
      }

      @Nullable
      public Icon getIcon(boolean open) {
        return myPresentation.getIcon(open);
      }

      @Nullable
      public TextAttributesKey getTextAttributesKey() {
        return myPresentation.getTextAttributesKey();
      }

      public String getName() {
        return myItem.getName();
      }

      public ItemPresentation getPresentation() {
        return myPresentation != null ? this : null;
      }

      public FileStatus getFileStatus() {
        return myItem.getFileStatus();
      }

      public void navigate(boolean requestFocus) {
        myItem.navigate(requestFocus);
      }

      public boolean canNavigate() {
        return myItem.canNavigate();
      }

      public boolean canNavigateToSource() {
        return myItem.canNavigateToSource();
      }

      public static void add(final NavigationItem item, Set<NavigationItem> symbolNavItems) {
        final ItemPresentation presentation;
        if (item instanceof PsiMetaOwner) {
          final PsiMetaData data = ((PsiMetaOwner)item).getMetaData();
          if (data instanceof PsiPresentableMetaData) {
            final PsiPresentableMetaData metaData = (PsiPresentableMetaData)data;
            presentation = new ItemPresentation() {
              public String getPresentableText() {
                return metaData.getName();
              }

              @Nullable
              public String getLocationString() {
                return MyNavigationItem.getLocationString((PsiElement)item);
              }

              @Nullable
              public Icon getIcon(boolean open) {
                return metaData.getIcon();
              }

              @Nullable
              public TextAttributesKey getTextAttributesKey() {
                final ItemPresentation p = item.getPresentation();
                return p != null ? p.getTextAttributesKey() : null;
              }
            };
          } else {
            presentation = item.getPresentation();
          }
        } else {
          presentation = item.getPresentation();
        }

        if (presentation != null) {
          symbolNavItems.add(new MyNavigationItem(item, presentation));
        }
      }
    }
  }
}
